/*
 *
 * mpuc.c  -  low level interface routines for Roland MPU-401 MIDI interface
 *
 * Interfaces to routines contained in assembly language routine MPUA.ASM
 * Borland C++
 *
 * Tests bidirectional communications with Chroma via MIDI
 *
 * Communication protocol of the 401:
 *   Initializing and resetting the 401
 *   Keeping track of 401 state so it can be reset correctly
 *   Installation of interrupt service routines on the PC
 *   Enabling, disabling, and servicing the interrupt controller
 *     chip in the PC
 *   Intercepting BDOS errors in PC dos
 *   Disabling the ctrl-break key
 *
 *
 */


#include <dos.h>


/* CLOCK DEFINES */

#define MASTER_CLOCK_RATE       96              /* 96 CLOCKS PER BEAT */
#define MIDI_CLOCKS_PER_BEAT    24

/* MISC INTERNAL DEFINES */

#define REC_CTRL_ALL      1
#define REC_CTRL_NO_BEND  0
#define REC_CTRL_NO_PRESS 2

#define COM_TYPE_401_COMMAND 0
#define COM_TYPE_401_READ 1
#define COM_TYPE_401_WRITE 2

#define SHELL_USE_INTERNAL_CLOCK 0
#define SHELL_USE_TAPE_CLOCK 1
#define SHELL_USE_MIDI_CLOCK 2

/* CODES FOR THE COMMANDS THAT ARE SENT TO THE 401 */

#define CT_401_MODE_PLAY_STOP          0x05
#define CT_401_MODE_PLAY_START         0x0A
#define CT_401_MODE_REC_STOP           0x11
#define CT_401_MODE_OVERDUB_STOP       0x15
#define CT_401_MODE_RECORD_STANDBY     0x20
#define CT_401_MODE_REC_START          0x22
#define CT_401_MODE_OVERDUB_START      0x2A

#define CT_401_NO_ALL_NOTES_OFF        0x30
#define CT_401_NO_REAL_TIME            0x32
#define CT_401_THRU_DEAD               0x33
#define CT_401_COMMON_TO_HOST          0x38
#define CT_401_REAL_TIME_IN            0x39
#define CT_401_UART                    0x3f
#define CT_401_SET_REFA                0x40
#define CT_401_USE_INTERNAL_CLOCK      0x80
#define CT_401_USE_TAPE_CLOCK          0x81
#define CT_401_USE_MIDI_CLOCK          0x82
#define CT_401_METRONOME_ON_NORMAL     0x83
#define CT_401_METRONOME_OFF           0x84
#define CT_401_METRONOME_ON_DOUBLE     0x85
#define CT_401_BENDER_OFF              0x86
#define CT_401_BENDER_ON               0x87
#define CT_401_THRU_OFF                0x88
#define CT_401_THRU_ON                 0x89
#define CT_401_INHIBIT_MEASURE_END     0x8C
#define CT_401_REQUEST_MEASURE_END     0x8D
#define CT_401_CONDUCTOR_OFF           0x8e
#define CT_401_CONDUCTOR_ON            0x8f
#define CT_401_RTM_AFFECT_OFF          0x90
#define CT_401_RTM_AFFECT_ON           0x91
#define CT_401_FSK_TO_INT              0x92
#define CT_401_FSK_TO_MIDI             0x93
#define CT_401_CLOCK_TO_HOST_OFF       0x94
#define CT_401_CLOCK_TO_HOST_ON        0x95
#define CT_401_SYS_EX_TO_HOST_OFF      0x96
#define CT_401_SYS_EX_TO_HOST_ON       0x97
#define CT_401_CREFA_OFF               0x98
#define CT_401_READ_RESET_RECORD_CTR   0xAB
#define CT_401_TEMPO_REQUEST           0xAF
#define CT_401_CLEAR_PLAY_COUNTERS     0xB8
#define CT_401_CLEAR_PLY               0xb9
#define CT_401_SET_CLOCK_RES           0xc2
#define CT_401_SEND_DATA               (0xd0 + 7)      /* send on track 7 */
#define CT_401_SEND_SYSTEM             0xdf
#define CT_401_SET_TEMPO               0xE0
#define CT_401_CLOCKS_PER_BEAT         0xE4
#define CT_401_BEATS_PER_MEASURE       0xE6
#define CT_401_CLOCK_TO_HOST_RATE      0xe7
#define CT_401_ACTIVE_TRACKS           0xEC
#define CT_401_ACCEPT_CHANNELS_LOW     0xEE
#define CT_401_ACCEPT_CHANNELS_HIGH    0xEF
#define CT_401_TIMING_OVERFLOW         0xF8
#define CT_401_NULL                    0xF8
#define CT_401_MEASURE_END             0xf9
#define CT_401_DATA_END                0xfc
#define CT_401_RESET                   0xFF

/* CODES FOR THE MESSAGES THAT ARE READ FROM THE 401 */

#define CF_401_REQUEST_PLAY_DATA       0xF0
#define CF_401_TIMING_OVERFLOW         0xF8
#define CF_401_DATA_END_MARK           0xFC
#define CF_401_ALL_END                 0xFC
#define CF_401_MEASURE_END             0xF9
#define CF_401_CONDUCT_REQ             0xf9
#define CF_401_CLOCK_TO_HOST           0xFD
#define CF_401_ACK                     0Xfe
#define CF_401_SYSTEM_MSG              0xff

/* HARDWARE LEVEL ACCESS TO THE 401, FROM 'C' */

#define BASE_ADDRESS_401       0X0330
#define STATUS_PORT_401        BASE_ADDRESS_401 + 1
#define COMMAND_PORT_401       BASE_ADDRESS_401 + 1
#define DATA_OUT_PORT_401      BASE_ADDRESS_401 + 0
#define DATA_IN_PORT_401       BASE_ADDRESS_401 + 0
#define DATA_READY_401_MASK    0x40
#define DATA_AVAILABLE_401_MASK 0x80


/* INPUT/OUTPUT LOW LEVEL MACROS */

#define NOT_READY_FOR_DATA_401() (inp(STATUS_PORT_401) & \
                                    DATA_READY_401_MASK)

#define COMMAND_OUT_401(com) outp(COMMAND_PORT_401, com)

#define DATA_OUT_401(dat) outp(DATA_OUT_PORT_401, dat)

#define NOT_DATA_AVAILABLE_401() (inp(STATUS_PORT_401) & \
                                    DATA_AVAILABLE_401_MASK)

#define READ_DATA_401() inp(DATA_IN_PORT_401)

/* STRUCTURE DEFINITION TO STORE SOME EXAMPLE GLOBAL
   SETUP INFORMATION
 */

typedef struct sh3
    {
    int tempo;                  /* in 1/256 * beats per minute */
    int record;                 /* true if record is enabled */
    int tsig_num;               /* time sig numerator */
    int tsig_den;               /* time sig denominator */
    int met_stat;               /* metronome enable */
	 int count_dn;               /* count down enable */
    int tape_sync;              /* tape synch is clock source */
    int rec_ctrl;               /* enable recording of midi controllers */
    int thru_stat;              /* midi thru is on for remote keyboard */
    int real_time_out;          /* enables xmission of real time commands
                                   if on */
    } SEQ_GLOBAL_TYPE;

SEQ_GLOBAL_TYPE status_globals;

/* FUNCTION DECLARATIONS */

/*int inport( int port );
void outport( int port, int word );
*/
void sh_update_tempo();
void update_met();
void shell_metronome_master();
void input_401();
void sh_send_command_401();
void sh_com_write_byte_401();
void sh_write_req_401();
void sh_command_401();
void init_status_globals();
void setup_401_state();
void re_init_401();
void system_exit();
void set_clock();
void shell_metronome_master();


/*
 * SH_SEND_COMMAND_401(COM)
 * sends a command to 401 with full handshake
 * does not read or write any extra bytes
 */

void sh_send_command_401(com)
int com;
{
sh_command_401(com, COM_TYPE_401_COMMAND);
}

/*
 * SH_COM_WRITE_BYTE_401(COM, BYTE)
 * sends command to 401 with handshake, then sends one byte of data
 */

void sh_com_write_byte_401(com, byte)
int com, byte;
{
char data = (char)byte;

sh_command_401(com, COM_TYPE_401_WRITE, &data, 1);
}

/*
 * SH_WRITE_REQ_401(COM, PTR, LENGTH)
 * sends a command to 401 with full handshake
 * then sends a data string after it.  ptr points to chars to go out,
 * length is number of bytes to send
 */

void sh_write_req_401(com, ptr, length)
int com, length;
char *ptr;
{
sh_command_401(com, COM_TYPE_401_WRITE, ptr, length);
}

/*
 * sh_command_401(COM, TYPE, PTR, LENGTH)
 * this is called by the three command passers above
 * it will send command to 401, then either return, wait for data or send
 * data
 */

void sh_command_401(com, type, ptr, length)
int com, type, length;
char *ptr;
{
int acked;
int x;

while ( NOT_READY_FOR_DATA_401());
disable_401();
COMMAND_OUT_401((char) com);           /* send the command to the 401 */

for(acked=0; !acked; )                  /* wait until our command is acked */
    {
	 x = read_401();                    /* get a byte back from the 401 */
    if (x == CF_401_ACK) acked++;      /* if command is an ack,
                                            we are done with command */
 /* else input_401(x);  */             /* deal with this input stream first */
    }
if (type == COM_TYPE_401_WRITE)        /* if write request... */
    for (; length; ptr++, length--) write_data_401(*ptr);
                                        /* write all the bytes to the 401 */
enable_401();                          /* ints restored to where they were
                                           before we shut them off */
}


/*
 * INITIALIZE_SHELL()
 * should be called once
 * to set up the shell before shell routines are called
 *
 * Returns - True if 401 reset ok
 */
initialize_shell()
{
int retval;

retval = setup_401();                          /* init 401 and interrupts */
if (retval)                                     /* if can init 401 */
    {
    init_status_globals();                      /* set sempo, etc... */
    setup_401_state();                         /* set 401 registers */
    }
return(retval);
}

/*  INIT_STATUS_GLOBALS()
 *
 * inits the Seq+ global vars
 */

void init_status_globals()
{
status_globals.tempo = 120;
status_globals.count_dn = 0;
status_globals.tape_sync = 0;
status_globals.rec_ctrl = REC_CTRL_NO_PRESS;
status_globals.real_time_out = 1;
status_globals.tsig_num = 4;
status_globals.tsig_den = 4;
}

/*
 * SETUP_401_STATE()
 * 4/4/85
 * sets up the 401 to reflect the current status globals.
 *  Must be done after every reset
 *  401 must be reset before this is called
 */

void setup_401_state()
{
sh_update_tempo(status_globals.tempo);
sh_send_command_401((status_globals.thru_stat) ?
                CT_401_THRU_ON :
                CT_401_THRU_DEAD);

if (!status_globals.thru_stat) sh_send_command_401(CT_401_THRU_OFF);
if (!status_globals.real_time_out) sh_send_command_401(CT_401_NO_REAL_TIME);

set_clock();                    /* set clock source */
shell_metronome_master();       /* set metronome time base */

sh_send_command_401((status_globals.rec_ctrl == REC_CTRL_NO_BEND) ?
                CT_401_BENDER_OFF :
					 CT_401_BENDER_ON);
                                /* setup the record midi controllers */

sh_send_command_401 (CT_401_RTM_AFFECT_OFF);
                                /* disable real time in */
sh_send_command_401(CT_401_REQUEST_MEASURE_END);
                                /* make 401 send end of measure */
sh_com_write_byte_401(CT_401_CLOCK_TO_HOST_RATE,
    (((MASTER_CLOCK_RATE * 4) /16) * 4)
                        );      /* set the rate */
sh_send_command_401(CT_401_NO_ALL_NOTES_OFF);
                                /* block all notes off commands */
}

/*
 * RE_INIT_401()
 * resets 401 and brings status up to date
 * used to change the "permanent" 401 funtions (real time, midi thru, ...),
 * or to make sure 401 is still in correct state.
 */

void re_init_401()
{
disable_401();                 /* no ints */
reset_401();                   /* hard reset */
setup_401_state();             /* initialize */
enable_401();                  /* restore ints */
}

/*
 * SETUP_401()
 * resets 401, sets up int vectors and enables int system
 *
 * Returns - true if 401 set ok
 */

static setup_401()
{
int retval;
set_int_vector();               /* set the interrupt vectors for service
                                   routines */
sh_init_enable_401();          /* turn on interrupt system for first time */
retval = reset_401();          /* try to reset 401 */
if (!retval) retval = reset_401();
                                /* if no good, try again */
return(retval);
}

/*
 * RESET_401()
 * attempts to do a low level software reset on the 401
 *
 * returns - true if success
 * (if failure, it may work if you try again
 */

reset_401()
{
unsigned x = 0;
int flag;
int ack_count, time_count;
int retval=1;                   /* assume success */

disable_401();                 /* disable 401 interrupts while
											  trying to reset */

READ_DATA_401();               /* Read a dummy byte to allow reset
											  if byte waiting */

for (time_count=5000, flag=1; time_count && flag; time_count--)
										  /* Here we will loop until ready for
											  data, or we give up  */
	 {
	 if (!(NOT_READY_FOR_DATA_401())) flag=0;
										  /* if 401 ready for data, we can quit loop */
	 }
if (flag)                       /* if we were timed out before we got ready
											  for data... */
	 {
	 retval = 0;                 /* ...then return flag couldn't reset */
	 }

else
	 {
	 COMMAND_OUT_401(CT_401_RESET);
										  /* Send the reset command */
	 for (time_count=10000, ack_count=5, flag=0; !flag;)
		  {                       /* loop here till time out, or ack recvd */
		  if ((NOT_DATA_AVAILABLE_401()))
										  /* if there is not data from 401 */
				{
				time_count--;       /* count one more wait */
				if (!time_count)
					 {
BAD_RESET:
					 flag++;         /* if time out, note it */
					 retval=0;       /* and note failure */
					 }
				}
		  else                    /* If there is data coming back form 401 */
				{
				x = (unsigned) READ_DATA_401();
										  /* ...read it */
				if (x == CF_401_ACK)
										  /* If data is the ACK message */
					 {
					 flag++;         /* note we are done, and no failure */
					 }
				else                /* if data back from 401 wasn't ack */
					 {
					 ack_count--;    /* note that we got something that isn't ack */
					 if (!ack_count) goto BAD_RESET;
										  /* If we get too many data bytes back from
											  401 with no ack, give up */
					 }
				}
		  }
	 }
enable_401();                  /* turn ints back on */
return(retval);
}

/*
 * SYSTEM_EXIT()
 * called to leave program.  resets ints and exits
 */

void system_exit()
{
extern void exit();

reset_401();
disable_401();
reset_int_vector();             /* put all the int vectors back where they
                                   were */
exit(1);
}

/*
 * SH_UPDATE_TEMPO(TEMPO)
 * sets 401 for new tempo.
 */

void sh_update_tempo(tempo)
int tempo;
{
sh_com_write_byte_401(CT_401_SET_TEMPO, tempo);
}


/*
 * FATAL_DOS_HANDLER(AX, DI)
 *
 * Dos fatal errors are caught by asm lang function, then the relevent
 * register contents are passed here for handling
 */

void fatal_dos_handler(ax, di)
{
printf("got fatal dos error, ax = %x, di = %x\n", ax, di);
        /* this is of course not a very usefull handler, but it shows
           how to intercept fatal dos errors
       */
}

/*
 * SET_CLOCK()
 *
 * Sets 401 for current sync source, timebase,
 */

void set_clock()
{
sh_send_command_401(CT_401_SET_CLOCK_RES + 2);
                            /* send command to set up internal resolution
                               to 96 clicks */
sh_send_command_401(CT_401_USE_INTERNAL_CLOCK + status_globals.tape_sync);
                            /* set clock source. Note that status.globals.
    tape_sync can only have the values SHELL_USE_INTERNAL_CLOCK to
    SHELL USE_MIDI_CLOCK (0..2) */
}

/*
 * SHELL_METRONOME_MASTER()
 * sets up 401 to reflect current time signature
 */

void shell_metronome_master()
{
update_met( status_globals.tsig_num, status_globals.tsig_den);
}

void update_met(num, den)
int num, den;
{
sh_com_write_byte_401(CT_401_CLOCKS_PER_BEAT,
                        ((MIDI_CLOCKS_PER_BEAT * 4) / den));
sh_com_write_byte_401(CT_401_BEATS_PER_MEASURE, num);
}
